#include "iMS_EnhancedSequences.h"
#include "MySequences.h"
#include "Supervisors.h"
#include "curses.h"

#include <sstream>

// Four window areas of the screen
WINDOW* command_win;
WINDOW* input_win;
WINDOW* queue_win;
WINDOW* status_win;

const char* Title = " iMS Enhanced Sequences Demo ";

const char* CommandList[][9] =
{
	// Main Menu
	{
	 "MAIN MENU",
	 "Create Sequence",
	 "Move Sequence",
	 "Delete Sequence",
	 "Update Termination",
	 "Start Playback (Int)",
	 "Start Playback (Ext)",
	 "Toggle Display Entries",
	 ""
	},

	{
	"PLAYBACK MENU",
	"Pause Sequence Queue (Gracefully)",
	"Pause Sequence Queue (Immediately)",
	"Resume Sequence Queue",
	"Stop Sequence Queue (Gracefully)",
	"Stop Sequence Queue (Immediately)",
	"Stop at End of Sequence",
	"",
	""
	},

	{
	"TONE PLAY MENU",
	"Move to next Sequence Entry",
	"Select New Tone",
	"",
	"",
	"",
	"",
	"",
	""
	},

	{
	"CREATE SEQUENCE MENU",
	"Add Image Entry to Sequence",
	"Add Tone Entry to Sequence",
	"Apply X axis offset",
	"Apply Y axis offset",
	"Finish Sequence and Download",
	"",
	"",
	""
	}

};

static int menu = 0;
static bool display_entries = false;

enum class SequenceStatus {
	STOPPED,
	PLAYING,
	PAUSED,
	TONEBUF
};

const char* SequenceStatusString[] = {
	"Stopped",
	"Image Playing",
	"Paused",
	"Tone Buffer Active"
};

static SequenceStatus status;

/// Displays the available commands for the current menu
static void PrintCommandMenu()
{
	int i;
	wclear(command_win);

	box(command_win, 0, 0);
	mvwprintw(command_win, 0, 2, " Command: ");
	mvwprintw(command_win, 1, 16, ">>>>> %s <<<<<", CommandList[menu][0]);
	mvwprintw(command_win, 0, COLS - 28, " >> [F12] = QUIT << ");

	for (i = 0; i < 8; i++) {
		mvwprintw(command_win, 2 + (i % 4), 5 + (COLS - 10) * (i / 4) / 2, "F%d: %s", i + 1, CommandList[menu][i+1]);
	}

	wrefresh(command_win);
}

static void HighlightMenuOption(int opt) {
	mvwchgat(command_win, 2 + (opt % 4), 4 + (COLS - 10) * (opt / 4) / 2, (COLS - 16) / 2, A_REVERSE, 0, NULL);
	wrefresh(command_win);
}

/// User input: Type a string
static std::string GetInputOption(const std::string& in)
{
	// Get string input
	nocbreak();
	echo();
	curs_set(1);

	char buf[256];
	memset(buf, '\0', sizeof(buf));

	wclear(input_win);
	box(input_win, 0, 0);
	mvwprintw(input_win, 0, 2, " Input: ");
	mvwprintw(input_win, 1, 3, in.c_str());
	mvwchgat(input_win, 1, COLS / 2, (COLS / 2) - 3, A_REVERSE, 0, NULL);
	wrefresh(input_win);
	wattrset(input_win, A_REVERSE);
	mvwgetstr(input_win, 1, (COLS / 2) + 1, buf);
	wattrset(input_win, A_NORMAL);

	curs_set(0);
	noecho();
	cbreak();

	return std::string(buf);
}

/// User Input: Select a Sequence
static ImageSequence& GetSequence(std::string in)
{
	int ch, index=0, count=myseq->Count();

	// Get termination action input
	wclear(input_win);
	box(input_win, 0, 0);
	mvwprintw(input_win, 0, 2, " Input: ");
	mvwprintw(input_win, 1, 3, in.c_str());
	mvwchgat(input_win, 1, COLS / 2, (COLS / 2) - 3, A_REVERSE, 0, NULL);
	wrefresh(input_win);
	wattrset(input_win, A_REVERSE);
	do {
		wmove(input_win, 1, (COLS / 2));
		wclrtoeol(input_win);
		mvwprintw(input_win, 1, (COLS / 2) + 1, "%s", myseq->FindSequence(index).Name().c_str());
		wrefresh(input_win);
		ch = getch();
		if (ch == KEY_UP || ch == KEY_LEFT) {
			index--;
			if (index < 0) index = count - 1;
		}
		else if (ch == KEY_DOWN || ch == KEY_RIGHT) {
			index++;
			if (index >= count) index = 0;
		}
	} while (ch != 0xA); // Enter
	wattrset(input_win, A_NORMAL);

	return myseq->FindSequence(index);
}

/// User Input: Termination Action
static SequenceTermAction GetTermAction()
{
	SequenceTermAction action = SequenceTermAction::RECYCLE;
	int ch;

	// Get termination action input
	wclear(input_win);
	box(input_win, 0, 0);
	mvwprintw(input_win, 0, 2, " Input: ");
	mvwprintw(input_win, 1, 3, "Choose a Termination Action (Up/Down):");
	mvwchgat(input_win, 1, COLS / 2, (COLS / 2) - 3, A_REVERSE, 0, NULL);
	wrefresh(input_win);
	wattrset(input_win, A_REVERSE);
	do {
		wmove(input_win, 1, (COLS / 2));
		wclrtoeol(input_win);
		mvwprintw(input_win, 1, (COLS / 2) + 1, "%s", myseq->SequenceTermString(action).c_str());
		wrefresh(input_win);
		ch = getch();
		if (ch == KEY_UP || ch == KEY_LEFT) {
			switch (action) {
			case SequenceTermAction::DISCARD: action = SequenceTermAction::REPEAT_FROM; break;
			case SequenceTermAction::STOP_DISCARD: action = SequenceTermAction::DISCARD; break;
			case SequenceTermAction::RECYCLE: action = SequenceTermAction::STOP_DISCARD; break;
			case SequenceTermAction::STOP_RECYCLE: action = SequenceTermAction::RECYCLE; break;
			case SequenceTermAction::INSERT: action = SequenceTermAction::STOP_RECYCLE; break;
			case SequenceTermAction::STOP_INSERT: action = SequenceTermAction::INSERT; break;
			case SequenceTermAction::REPEAT: action = SequenceTermAction::STOP_INSERT; break;
			case SequenceTermAction::REPEAT_FROM: action = SequenceTermAction::REPEAT; break;
			default: action = SequenceTermAction::RECYCLE; break;
			}
		}
		else if (ch == KEY_DOWN || ch == KEY_RIGHT) {
			switch (action) {
			case SequenceTermAction::DISCARD: action = SequenceTermAction::STOP_DISCARD; break;
			case SequenceTermAction::STOP_DISCARD: action = SequenceTermAction::RECYCLE; break;
			case SequenceTermAction::RECYCLE: action = SequenceTermAction::STOP_RECYCLE; break;
			case SequenceTermAction::STOP_RECYCLE: action = SequenceTermAction::INSERT; break;
			case SequenceTermAction::INSERT: action = SequenceTermAction::STOP_INSERT; break;
			case SequenceTermAction::STOP_INSERT: action = SequenceTermAction::REPEAT; break;
			case SequenceTermAction::REPEAT: action = SequenceTermAction::REPEAT_FROM; break;
			case SequenceTermAction::REPEAT_FROM: action = SequenceTermAction::DISCARD; break;
			default: action = SequenceTermAction::RECYCLE; break;
			}
		}
	} while (ch != 0xA); // Enter
	wattrset(input_win, A_NORMAL);

	return action;
}

/// Intermediate object for creating a sequence (obtain from MySequences)
static ImageSequence* new_sequence = nullptr;

/// This is called to perform an action based on the user selection of a menu option
static int DoMenuCommand(int ch)
{
	static int seq_count = 0;
	static std::shared_ptr<SequenceEntry> entry_ref;

	if (ch == KEY_F(12)) return -1;

	int opt = (ch - KEY_F(1));
	if (ch >= KEY_F(1) && ch <= KEY_F(8) && (strlen(CommandList[menu][opt+1]) > 0))
		HighlightMenuOption(opt);

	switch (menu) {
		// Main Menu
	case 0: {
		switch (ch) {
		case KEY_F(1): {
			std::string str = GetInputOption("Enter Sequence Name:");
			if (str.size() == 0) {
				std::stringstream ss;
				ss << "Example Sequence " << seq_count++;
				str = ss.str();
			}
			new_sequence = myseq->CreateSequence();
			new_sequence->Name() = str;
			menu = 3;
		}
			break;
		case KEY_F(2): {
			auto& seq = GetSequence("Move which sequence?");
			auto& seq2 = GetSequence("Move before which other sequence?");
			glbl->seqMgr().MoveSequence(seq2, seq);
		}
			break;
		case KEY_F(3): {
			auto& seq = GetSequence("Delete which sequence?");
			glbl->seqMgr().RemoveSequence(seq);
		}
			break;
		case KEY_F(4): {
			std::stringstream ss;
			int count = glbl->seqMgr().QueueCount(), which = 0;
			ss << "Which Sequence in Queue? (0=HEAD, " << count - 1 << "=TAIL)";
			std::string id = GetInputOption(ss.str());
			try {
				which = std::stoi(id);
			}
			catch (...) {
				break;
			}
			std::array<uint8_t, 16> uuid;
			glbl->seqMgr().GetSequenceUUID(which, uuid);
			auto action = GetTermAction();
			if ((action == SequenceTermAction::INSERT) || (action == SequenceTermAction::STOP_INSERT)) {
				ImageSequence& seq = GetSequence("Insert before which sequence?");
				glbl->seqMgr().UpdateTermination(uuid, action, seq.GetUUID());
				myseq->FindSequence(uuid).OnTermination(action, &seq);
			}
			else if (action == SequenceTermAction::REPEAT_FROM) {
				std::string str = GetInputOption("Repeat from which Entry Index:");
				int idx = 0;
				try {
					idx = std::stoi(str);
				}
				catch (...) {
					idx = 0;
				}
				glbl->seqMgr().UpdateTermination(uuid, action, idx);
				myseq->FindSequence(uuid).OnTermination(action, idx);
			}
			else {
				glbl->seqMgr().UpdateTermination(uuid, action);
				myseq->FindSequence(uuid).OnTermination(action);
			}
		}
			break;
		case KEY_F(5): {
			SequenceManager::SeqConfiguration cfg (
				SequenceManager::PointClock::INTERNAL,
				SequenceManager::ImageTrigger::CONTINUOUS
				);
			glbl->seqMgr().StartSequenceQueue(cfg,
				SequenceManager::ImageTrigger::CONTINUOUS);
		}
			break;
		case KEY_F(6): {
			SequenceManager::SeqConfiguration cfg (
				SequenceManager::PointClock::EXTERNAL,
				SequenceManager::ImageTrigger::EXTERNAL
				);
			glbl->seqMgr().StartSequenceQueue(cfg,
				SequenceManager::ImageTrigger::EXTERNAL);
		}
			break;
		case KEY_F(7): 
			display_entries = !display_entries;
			break;
		default: 
			break;
		}
		break;
	}
		  // Playing Menu
	case 1: {
		switch (ch) {
		case KEY_F(1):
			glbl->seqMgr().Pause(ImagePlayer::StopStyle::GRACEFULLY);
			status = SequenceStatus::PAUSED;
			break;
		case KEY_F(2):
			glbl->seqMgr().Pause(ImagePlayer::StopStyle::IMMEDIATELY);
			status = SequenceStatus::PAUSED;
			break;
		case KEY_F(3):
			glbl->seqMgr().Resume();
			status = SequenceStatus::PLAYING;
			break;
		case KEY_F(4):
			glbl->seqMgr().Stop(ImagePlayer::StopStyle::GRACEFULLY);
			break;
		case KEY_F(5):
			glbl->seqMgr().Stop(ImagePlayer::StopStyle::IMMEDIATELY);
			break;
		case KEY_F(6):
			glbl->seqMgr().StopAtEndOfSequence();
			break;
		default:
			break;
		}
		break;
	}
		  // Tone Buffer Menu
	case 2: {
		switch (ch) {
		case KEY_F(1): {
			glbl->seqMgr().SendHostTrigger();
			status = SequenceStatus::PLAYING;
		}
			break;
		case KEY_F(2): {
			std::string str = GetInputOption("Select a Tone Index (0-255):");
			int idx = 0;
			try {
				idx = std::stoi(str);
			}
			catch (...) {
				idx = 0;
			}
			if (idx < 0) idx = 0;
			if (idx > 255) idx = 255;

			glbl->sigpath().UpdateLocalToneBuffer(idx);
		}
			break;
		case KEY_F(3):
			break;
		case KEY_F(4):
			break;
		case KEY_F(5):
			break;
		case KEY_F(6):
			break;
		default:
			break;
		}
		break;
	}
		  // Create Sequence
	case 3: {
		if (new_sequence == nullptr) {
			menu = 1;
			break;
		}
		switch (ch) {
		case KEY_F(1): {
			// Add Image Sequence Entry
			std::string str = GetInputOption("Image Index (default = random selection):");
			int idx = -1, rpts = 100;
			if (str.size() > 0) {
				try {
					idx = std::stoi(str);
				}
				catch (...) {
					idx = -1;
				}
			}
			const Image& img = getImage(idx);
			std::string str2 = GetInputOption("Number of Repeats (default = 100):");
			if (str2.size() > 0) {
				try {
					rpts = std::stoi(str2);
				}
				catch (...) {
					rpts = 100;
				}
				if (rpts < 0) rpts = 0;
			}
			auto entry = std::make_shared<ImageSequenceEntry>(img, ImageRepeats::PROGRAM, rpts);
			entry_ref = entry;  // keep a reference to the entry for applying offsets
			new_sequence->push_back(entry);
		}
			break;
		case KEY_F(2): {
			auto entry = std::make_shared<ToneSequenceEntry>(getToneMap());
			entry_ref = entry;
			new_sequence->push_back(entry);
		}
			break;
		case KEY_F(3): {
			std::string str = GetInputOption("Enter X Axis Frequency Offset (MHz):");
			double d = 0.0;
			try {
				d = std::stod(str);
			}
			catch (...) {
				d = 0.0;
			}
			entry_ref->SetFrequencyOffset(MHz(d), RFChannel(1));
			entry_ref->SetFrequencyOffset(MHz(d), RFChannel(2));
		}
			break;
		case KEY_F(4): {
			std::string str = GetInputOption("Enter Y Axis Frequency Offset (MHz):");
			double d = 0.0;
			try {
				d = std::stod(str);
			}
			catch (...) {
				d = 0.0;
			}
			entry_ref->SetFrequencyOffset(MHz(d), RFChannel(3));
			entry_ref->SetFrequencyOffset(MHz(d), RFChannel(4));
		}
			break;
		case KEY_F(5): {
			auto action = GetTermAction();
			if ((action == SequenceTermAction::INSERT) || (action == SequenceTermAction::STOP_INSERT)) {
				auto& seq = GetSequence("Insert before which sequence?");
				new_sequence->OnTermination(action, &seq);
			}
			else if (action == SequenceTermAction::REPEAT_FROM) {
				std::string str = GetInputOption("Repeat from which Entry Index:");
				int idx = 0;
				try {
					idx = std::stoi(str);
				}
				catch (...) {
					idx = 0;
				}
				new_sequence->OnTermination(action, idx);
			}
			else {
				new_sequence->OnTermination(action);
			}
			myseq->DownloadSequence(*new_sequence);
			new_sequence = nullptr;
			menu = 0;
		}
			break;
		case KEY_F(6):
			break;
		default:
			break;
		}
		break;
	}
	default: break;
	}
	wrefresh(input_win);
	return 0;
}

/// Prints information in the middle of the screen about the contents of the sequence queue (or the new sequence if in the process of creating a new sequence)
void DrawQueue()
{
	wclear(queue_win);

	box(queue_win, 0, 0);

	if (new_sequence != nullptr) {
		mvwprintw(queue_win, 0, 2, " New Sequence: ");
		int queue_pos = 2;
		int queue_end = getmaxy(queue_win) - 2;
		mvwprintw(queue_win, queue_pos, 3, " --> %s", new_sequence->Name().c_str());
		mvwprintw(queue_win, queue_pos, 50, "-- %4d entries  --  Term: %s", new_sequence->size(), myseq->SequenceTermString(new_sequence->TermAction()).c_str());
		queue_pos++;
		for (ImageSequence::const_iterator it = new_sequence->begin(); it != new_sequence->end(); it++) {
			if (queue_pos >= queue_end) {
				mvwprintw(queue_win, queue_pos, 10, ". . . . . continued . . . . .");
				break;
			}
			if (matchesToneBuffer((*it)->UUID())) {
				mvwprintw(queue_win, queue_pos, 6, "> Tone Buffer Entry <");
			}
			else {
				auto& img = getImage((*it)->UUID());
				mvwprintw(queue_win, queue_pos, 6, "> Image \"%s\"", img.Name().c_str());
				mvwprintw(queue_win, queue_pos, 50, "-- %6d pts", img.Size());
			}
			double f1 = (*it)->GetFrequencyOffset(RFChannel(1));
			double f2 = (*it)->GetFrequencyOffset(RFChannel(2));
			double f3 = (*it)->GetFrequencyOffset(RFChannel(3));
			double f4 = (*it)->GetFrequencyOffset(RFChannel(4));
			mvwprintw(queue_win, queue_pos, 67, "--  offsets = [ %.2f, %.2f, %.2f, %.2f] MHz",
				f1, f2, f3, f4);
			mvwprintw(queue_win, queue_pos, 118, "-- repeats = %d", (*it)->NumRpts());
			queue_pos++;
		}
	}
	else {
		mvwprintw(queue_win, 0, 2, " Sequence Queue: ");

		if ((status == SequenceStatus::STOPPED) || (status == SequenceStatus::PAUSED)) {
			int count = glbl->seqMgr().QueueCount();

			mvwprintw(queue_win, 2, 3, "Total Sequences in queue: %d", count);

			int queue_pos = 4;
			int queue_end = getmaxy(queue_win) - 2;
			for (int i = 0; i < count; i++) {
				if (queue_pos >= queue_end) {
					mvwprintw(queue_win, queue_pos, 10, ". . . . . continued . . . . .");
					break;
				}

				std::array<uint8_t, 16> uuid;
				glbl->seqMgr().GetSequenceUUID(i, uuid);
				auto& seq = myseq->FindSequence(uuid);
				if (i == 0) {
					mvwprintw(queue_win, queue_pos, 3, "[HEAD]: %s", seq.Name().c_str());
				}
				else if (i == (count - 1)) {
					mvwprintw(queue_win, queue_pos, 3, "[TAIL]: %s", seq.Name().c_str());
				}
				else {
					mvwprintw(queue_win, queue_pos, 3, "[%04d]: %s", i, seq.Name().c_str());
				}
				mvwprintw(queue_win, queue_pos, 50, "-- %4d entries  --  Term: %s", seq.size(), myseq->SequenceTermString(seq.TermAction()).c_str());
				queue_pos++;
				if (display_entries) {
					for (ImageSequence::const_iterator it = seq.begin(); it != seq.end(); it++) {
						if (queue_pos >= queue_end) {
							mvwprintw(queue_win, queue_pos, 10, ". . . . . continued . . . . .");
							break;
						}
						if (matchesToneBuffer((*it)->UUID())) {
							mvwprintw(queue_win, queue_pos, 6, "> Tone Buffer Entry <");
						}
						else {
							auto& img = getImage((*it)->UUID());
							mvwprintw(queue_win, queue_pos, 6, "> Image \"%s\"", img.Name().c_str());
							mvwprintw(queue_win, queue_pos, 50, "-- %6d pts", img.Size());
						}
						double f1 = (*it)->GetFrequencyOffset(RFChannel(1));
						double f2 = (*it)->GetFrequencyOffset(RFChannel(2));
						double f3 = (*it)->GetFrequencyOffset(RFChannel(3));
						double f4 = (*it)->GetFrequencyOffset(RFChannel(4));
						mvwprintw(queue_win, queue_pos, 67, "--  offsets = [ %.2f, %.2f, %.2f, %.2f] MHz",
							f1, f2, f3, f4);
						mvwprintw(queue_win, queue_pos, 118, "-- repeats = %d", (*it)->NumRpts());
						queue_pos++;
					}
				}
			}
		}
		else {
			mvwprintw(queue_win, 2, 3, "Cannot get status information while playing");
		}
	}
	wrefresh(queue_win);
}

/// Prints information at the bottom of the screen about current status
void DrawStatus(SequenceSupervisor& sup)
{
	static std::string SeqPlaying = "";
	int position = -1;

	wclear(status_win);
	box(status_win, 0, 0);
	mvwprintw(status_win, 0, 2, " Status: ");

	mvwprintw(status_win, 1, 15, " >>> %s <<<", SequenceStatusString[(int)status]);

	if (sup.SequenceStarted()) {
		SeqPlaying = myseq->FindSequence(sup.UUID()).Name();
		mvwprintw(status_win, 1, 60, "Started Sequence %s", SeqPlaying.c_str());
	}
	if ((position = sup.GetPosition()) >= 0) {
		SeqPlaying = myseq->FindSequence(sup.UUID()).Name();
		mvwprintw(status_win, 1, 60, "Current Position: Seq [%s]  ===>  Entry # [%d]", SeqPlaying.c_str(), position);
	}

	// Request position
	if ((status == SequenceStatus::PLAYING) || (status == SequenceStatus::PAUSED)) {
		glbl->seqMgr().GetCurrentPosition();
	}

	wrefresh(status_win);
}

/// Run this to move between different modes depending on the status of the iMS sequencer
void UpdateStatus(SequenceSupervisor& sup)
{
	int last_menu = menu;

	switch (status) {
	case SequenceStatus::PAUSED: 
		menu = 1;
		break;
	case SequenceStatus::STOPPED: 
		if (sup.Playing()) {
			status = SequenceStatus::PLAYING;
			menu = 1;
		}
		break;
	case SequenceStatus::PLAYING: 
		if (!sup.Playing()) {
			status = SequenceStatus::STOPPED;
			menu = 0;
		}
		else {
			if (sup.ToneStarted()) {
				status = SequenceStatus::TONEBUF;
				menu = 2;
			} else menu = 1;
		}
		break;
	case SequenceStatus::TONEBUF: 
		menu = 2;
		break;
	}

	//if (last_menu != menu) {
		PrintCommandMenu();
	//}
}

/// Call this regularly to perform actions based on user input
int RunMenu() 
{
	timeout(500);
	int ch = getch();
	timeout(-1);

	// Reset User input box
	wclear(input_win);
	box(input_win, 0, 0);
	mvwprintw(input_win, 0, 2, " Input: ");
	wrefresh(input_win);

	if (ch != ERR) {
		if (-1 == DoMenuCommand(ch))
			return -1;
	}
	return 0;
}

/// Run this once to initialise and display the menu
void SequenceMenuInit()
{
	status = SequenceStatus::STOPPED;

	initscr();
	resize_term(60, 150);
	cbreak();  // Line buffering disabled
	noecho();
	curs_set(0);
	
	command_win = newwin(8, COLS - 2, 1, 1);
	input_win = newwin(3, COLS - 2, 9, 1);
	queue_win = newwin(LINES - 15, COLS - 2, 12, 1);
	status_win = newwin(3, COLS - 2, LINES - 3, 1);

	keypad(stdscr, TRUE);  // Get function keys

						   // Draw Title Bar
	mvhline(0, 1, 0, COLS - 2);
	mvprintw(0, (COLS - strlen(Title)) / 2, "%s", Title);

	PrintCommandMenu();
	 
	box(input_win, 0, 0);
	mvwprintw(input_win, 0, 2, " Input: ");
	wrefresh(input_win);

	refresh();
}

/// Run this when finished with the menu system
void SequenceMenuDeinit()
{
	delwin(command_win);
	delwin(input_win);
	delwin(queue_win);
	delwin(status_win);

	curs_set(1);
	nocbreak();
	echo();
	endwin();
}